// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package event;

import org.omg.CORBA.*;
import org.omg.PortableServer.*;
import org.omg.CosTypedEventChannelAdmin.*;
import org.omg.CosEventChannelAdmin.*;

public class TypedPushClient
{
    static int
    run(ORB orb, String args[])
        throws UserException
    {
        //
        // Resolve Root POA.
        //
        POA rootPOA =
            POAHelper.narrow(orb.resolve_initial_references("RootPOA"));

        //
        // Get a reference to the POA manager.
        //
        POAManager manager = rootPOA.the_POAManager();

        //
        // Parse command line arguments.
        //
        boolean slow = false;
        long timeout = 5;
        boolean reg = true;
        String channelName = null;
        for(int i = 0; i < args.length; i++)
        {
	    if(args[i].equals("--help") || args[i].equals("-h"))
	    {
	        usage();
                return 0;
	    }
	    else if(args[i].equals("-s") || args[i].equals("--slow"))
	    {
	        slow = true;
	    }
	    else if(args[i].equals("-n") ||args[i].equals("--no-register"))
	    {
	        reg = false;
	    }
	    else if(args[i].equals("-t") || args[i].equals("--timeout"))
	    {
	        if(i + 1 >= args.length)
	        {
		    System.err.println("Argument expected for " + args[i]);
		    return 1;
	        }
	        try
	        {
		    timeout = Long.parseLong(args[i+1]);
	        }
	        catch(NumberFormatException ex)
	        {
		    System.err.println("Bad format for timeout: " + args[i+1]);
		    return 1;
	        }
	        ++i;
	    }
	    else if(!args[i].startsWith("-"))
	    {
	        channelName = args[i];
	    }
	    else
	    {
	        System.err.println("Unknown option `" + args[i] + "'");
	        usage();
	        return 1;
	    }
        }
	    
        //
        // Get event channel.
        //
        TypedEventChannel e = GetTypedEventChannel(orb, channelName);

        //
        // Create a persistent POA.
        //
        Policy[] pl = new Policy[2];
        pl[0] = rootPOA.create_lifespan_policy(
	    org.omg.PortableServer.LifespanPolicyValue.PERSISTENT);
        pl[1] = rootPOA.create_id_assignment_policy(
	    org.omg.PortableServer.IdAssignmentPolicyValue.USER_ID);

        POA consumerPOA = rootPOA.create_POA("consumer", manager, pl);

        //
        // Create the servant
        //
        TimerUpdate_impl impl = new TimerUpdate_impl(orb, consumerPOA, slow);

        //
        // Connect the supplier to the POA with the same name each
        // time.
        //
        byte[] oid = ("DefaultConsumer").getBytes();
        consumerPOA.activate_object_with_id(oid, impl);
        org.omg.CosEventComm.PushConsumer consumer = impl._this(orb);

        //
        // Activate the POAManager.
        //
        manager.activate();

    	ProxyPushSupplier supplier = null;
        if(reg)
        {
	    //
	    // Connect implementation to TypedProxyPushSupplier
	    //
	    TypedConsumerAdmin consumerAdmin = e.for_consumers();
            supplier = consumerAdmin.obtain_typed_push_supplier(
                "IDL:event/TimerUpdate:1.0");
            supplier.connect_push_consumer(consumer);
        }

	ShutdownThread t = null;

	if(timeout > 0)
	{
	    //
	    // Create an instance of the ShutdownThread
	    //
	    t = new ShutdownThread(orb, timeout);

	    //
	    // Start a thread which will shutdown the ORB after timeout
	    // seconds
	    //
	    t.start();
	}

	//
	// Run implementation.
	//
	orb.run();

	//
	// If the shutdown thread exists then tell it to go away
	//
	if(t != null)
	{
	    t.destroy();
	    //
	    // Wait for the thread to terminate
	    //
	    while(t.isAlive())
	    {
		try
		{
		    t.join();
		}
		catch(InterruptedException ex)
		{
		}
	    }
	}

        try
        {
	    //
	    // Disconnect the push supplier here to stop the event
	    // service from pushing further events
	    //
	    if(reg)
	        supplier.disconnect_push_supplier();
	    orb.shutdown(true);
        }
        //
        // Expected if ORB/Client has already been shutdown
        //
        catch(TRANSIENT ex)
        {
        }
        catch(BAD_INV_ORDER ex)
        {
        }
        catch(INITIALIZE ex)
        {
        }
        catch(NO_RESPONSE ex)
        {
        }

	return 0;
    }

    public static void
    main(String args[])
    {
        java.util.Properties props = System.getProperties();
        props.put("org.omg.CORBA.ORBClass", "com.ooc.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass",
                  "com.ooc.CORBA.ORBSingleton");

        int status = 0;
        ORB orb = null;

        try
        {
            args = com.ooc.CORBA.ORB.ParseArgs(args, props, null);

	    //
	    // Calling disconnect_* can cause the event service to call
	    // back to inform the proxy that it is disconnecting
	    // it. Therefore it's necessary to allow nested method
	    // invocations.
	    //
	    props.put("ooc.orb.conc_model", "threaded");
	    props.put("ooc.orb.oa.conc_model", "threaded");

            org.omg.CORBA.StringSeqHolder argsH =
                new org.omg.CORBA.StringSeqHolder(args);
            orb = com.ooc.CORBA.ORB.init(argsH, props, null);
            status = run(orb, argsH.value);
        }
        catch(Exception ex)
        {
            ex.printStackTrace();
            status = 1;
        }

        if(orb != null)
        {
            try
            {
                orb.destroy();
            }
            catch(Exception ex)
            {
                ex.printStackTrace();
                status = 1;
            }
        }

        System.exit(status);
    }

    private static void
    usage()
    {
        System.out.println(
            "Usage: java event.TypedPushClient [options] [channel-name]\n\n" +
	    "Options:\n" +
	    "-h, --help             Show this message.\n" +
	    "-s, --slow             Run in `slow' mode.\n" +
	    "-t sec, --time sec     Run for [sec] seconds, -1 forever.\n" +
	    "-n, --no-register      Don't register with the event service.\n");
    }

    static private TypedEventChannel
    GetTypedEventChannel(org.omg.CORBA.ORB orb, String name)
    {
	TypedEventChannel e = null;
	if(name == null)
	{
	    //
	    // Get event channel from the initial reference.
	    //
	    try
	    {
		org.omg.CORBA.Object obj =
		    orb.resolve_initial_references("TypedEventService");
		e = TypedEventChannelHelper.narrow(obj);
	    }
	    catch(org.omg.CORBA.ORBPackage.InvalidName ex)
	    {
		System.err.println("Can't resolve `TypedEventService'");
		System.exit(1);
	    }
            catch(org.omg.CORBA.BAD_PARAM ex)
            {
                // narrow() failed
            }

	    if(e == null)
	    {
		System.err.println("`TypedEventService' is not a " +
				   "TypedEventChannel object reference");
		System.exit(1);
	    }
	}
	else
	{
	    //
	    // Get event channel factory. Note that this uses ORBacus
	    // proprietary methods.
	    //
	    com.ooc.OBTypedEventChannelFactory.TypedEventChannelFactory f
		= null;
	    try
	    {
		org.omg.CORBA.Object obj =
		    orb.resolve_initial_references("TypedEventChannelFactory");
		f = com.ooc.OBTypedEventChannelFactory.
		    TypedEventChannelFactoryHelper. narrow(obj);
	    }
	    catch(org.omg.CORBA.ORBPackage.InvalidName ex)
	    {
		System.err.println("Can't resolve `TypedEventChannelFactory'");
		System.exit(1);
	    }
            catch(org.omg.CORBA.BAD_PARAM ex)
            {
                // narrow() failed
            }
	
	    if(f == null)
	    {
		System.err.println("`TypedEventChannelFactory' is not a " +
				   "TypedEventChannelFactory object " +
				   "reference");
		System.exit(1);
	    }

	    //
	    // Get event channel.
	    //
	    try
	    {
		e = f.get_channel_by_id(name);
	    }
	    catch(com.ooc.OBEventChannelFactory.ChannelNotAvailable ex)
	    {
		try
		{
		    e = f.create_channel(name);
		    System.err.println("`Channel " + name + "' created");
		}
		catch(com.ooc.OBEventChannelFactory.ChannelAlreadyExists ex1)
		{
		    System.err.println("Channel exists: try again");
		    System.exit(1);
		}
	    }
	}
	return e;
    }
}
