// ********************************************************************* 
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
// 
// All Rights Reserved
//
// **********************************************************************

public class ICP
{
    static final short thermometer = 0;
    static final short thermostat = 1;

    static private class DeviceState            // State for a device
    {
	short  type;
	String model;
	String location;
	short  nominal_temp;                    // For thermostats only
    }

    //----------------------------------------------------------------

    static final short MIN_TEMP = 40;      // 40 F ==  4.44 C
    static final short MAX_TEMP = 90;      // 90 F == 32.22 C
    static final short DFLT_TEMP = 68;     // 68 F == 20.00 C

    // Map of known devices
    static private java.util.Hashtable dstate = new java.util.Hashtable();

    //----------------------------------------------------------------

    // online() simulates adding a new device to the network by
    // adding it to the dstate map.
    //
    // For this simple simulation, devices with odd asset numbers
    // are thermometers and devices with even asset numbers
    // are thermostats.
    //
    // Thermostats get an initial nominal temperature of DFLT_TEMP.
    // The location string is intentionally left blank because it
    // must be programmed by the controller after putting the device
    // on-line (as should be the nominal temperature).
    //
    // If a device with the specified ID is on-line already, the
    // return value is -1. A zero return value indicates success.

    static public int 
    online(int id)            // Add device
    {
	// Look for id in state map.
	Object objectId = new Integer(id);
	Object exists = (DeviceState)dstate.get(objectId);
	if(exists != null)
	    return -1;                          // Already exists

	// Fill in state.
	DeviceState ds = new DeviceState();
	if((id % 2) == 0)
	    ds.type = thermostat;
        else
	    ds.type = thermometer;

	ds.model = (ds.type == thermometer)
	    ? "Sens-A-Temp" : "Select-A-Temp";
	ds.nominal_temp = DFLT_TEMP;

	// Insert new device into map
	dstate.put(objectId, ds);

	return 0;
    }

    //----------------------------------------------------------------

    // offline() simulates removing a device from the network by
    // removing it from the dstate map. If the device isn't known, the
    // return value is -1. A zero return value indicates success.

    static public int 
    offline(int id)           // Remove device
    {
	// Look for id in state map
	Object objectId = new Integer(id);
	Object exists = (DeviceState)dstate.get(objectId);
	if(exists == null)
	    return -1;                          // No such device

	dstate.remove(objectId);

	return 0;
    }

    //----------------------------------------------------------------

    // get() returns an attribute value of the device with the
    // given id. The attribute is named by the attr parameter. 
    //
    // By default, thermometers report a temperature that varies
    // somewhat around DFLT_TEMP. However, if there is another
    // thermostat in the same room as the the thermometer, the
    // thermometer reports a temperature that varies around that
    // thermostat's temperature. For several thermostats that are in
    // the same room, the thermometer reports around the average
    // nominal temperature of all the thermostats.
    //
    // Attempts to read from a non-existent device or to read a
    // non-existent attribute return -1. A return value of zero
    // indicates success. 

    static public Object 
    get(int id, String attr)
    {
	// Look for id in state map
	Object objectId = new Integer(id);
	DeviceState posObj = (DeviceState)dstate.get(objectId);
	if(posObj == null)
	    return null;                        // No such device

	// Depending on the attribute, return the
	// corresponding piece of state.
	Object return_value = null;
	if(attr.equals("model") == true)
        {
	    return_value = posObj.model;
	}
        else if(attr.equals("location") == true)
        {
	    return_value = posObj.location;
	}
        else if(attr.equals("nominal_temp"))
        {
	    if(posObj.type != thermostat)
		return null;                    // Must be thermostat
	    return_value = new Short(posObj.nominal_temp);
	}
        else if(attr.equals("temperature"))
        {
	    return_value = new Short(actual_temp(posObj));
	}
        else if(attr.equals("MIN_TEMP") == true)
        {
	    return_value = new Short(MIN_TEMP);
	}
        else if(attr.equals("MAX_TEMP") == true)
        {
	    return_value = new Short(MAX_TEMP);
	}
        else
        {
	    return null;                        // No such attribute
	}

	return return_value;
    }

    //----------------------------------------------------------------

    // set() sets the the attributed specified by attr to the
    // value specified by value for the device with ID id. 
    // Attempts to access a non-existent device or attribute
    // return -1. Attempts to set a nominal temperature outside the
    // legal range also return -1. A zero return value
    // indicates success.

    static public int 
    set(int id, String attr, Object value)
    {
	// Look for id in state map
	Object objectId = new Integer(id);
	DeviceState posObj = (DeviceState)dstate.get(objectId);
	if(posObj == null)
	    return -1;                          // No such device

	// Change either location or nominal temp, depending on attr.
	if(attr.equals("location") == true)
        {
	    posObj.location = String.valueOf(value);
	}
        else if(attr.equals("nominal_temp") == true)
        {
	    if(posObj.type != thermostat)
		return -1;                      // Must be thermostat
	    short temp = Short.parseShort(value.toString());
	    if(temp < MIN_TEMP || temp > MAX_TEMP)
		return -1;
	    posObj.nominal_temp = temp;
	}
        else
        {
	    return -1;                          // No such attribute
	}
	return 0;                               // OK
    }

    //----------------------------------------------------------------

    // actual_temp() is a helper function to determine the actual
    // temperature returned by a particular thermometer or thermostat.
    // The pos argument indicates the device.
    //
    // The function locates all thermostats that are in the same room
    // as the device denoted by pos and computes the average of all
    // the thermostats' nominal temperatures. (If no thermostats are
    // in the same room as the device, the function assumes that the
    // average of the nominal temperatures is DFLT_TEMP.)
    //
    // The returned temperature varies from the average as
    // determined by vary_temp().

    static
    private short
    actual_temp(Object pos)
    {
	long sum = 0;
	long count = 0;

	// Find the first occurance of an object with same room
	// temperataure as that of the inputer parameter posObj.
	java.util.Enumeration e = dstate.elements();
	while(e.hasMoreElements())
        {
	    DeviceState ds = (DeviceState)e.nextElement();
	    if(ds.type == thermostat && 
               ds.location.equals(((DeviceState)pos).location))
            {
	        count++;
	        sum += ds.nominal_temp;
	    }
	}

        return vary_temp(count == 0 ? DFLT_TEMP : (short)(sum / count));
    }

    //----------------------------------------------------------------

    // vary_temp() simulates the variation in actual temperature
    // around a thermostat.The function randomly varies the
    // temperature as a percentage of calls as follows:
    //
    //      3 degrees too cold:      5%
    //      3 degrees too hot:       5%
    //      2 degrees too cold:     10%
    //      2 degrees too hot:      10%
    //      1 degree too cold:      15%
    //      1 degree too hot:       15%
    //      exact temperature:      40%

    static
    private short
    vary_temp(short temp)
    {
	java.util.Random rand = new java.util.Random();
	long r = rand.nextLong() % 50;
	short delta;
	if(r < 5)
	    delta = 3;
	else if(r < 15)
	    delta = 2;
	else if(r < 30)
	    delta = 1;
	else
	    delta = 0;
	r = rand.nextLong() % 2;
	if(r != 0)
	    delta = (short)-delta;

	return (short)(temp + delta);
    }
}
