// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package factory;

public class Client
{
    //
    // Test basic factory operations
    //
    static void
    TestFactory(DBFactory factory)
    {
        //
        // Purge any records left in the database
        //
        {
            int[] records = factory.getAllKeys();
            System.out.println("Purging " + records.length + " records.");
            for(int i = 0 ; i < records.length ; i++)
            {
                try
                {
                    factory.destroyByKey(records[i]);
                }
                catch(factory.DBFactoryPackage.NoRecord ex)
                {
                    System.out.println("Record " + records[i] +
                                       " doesn't exist anymore.");
                }
            }
        }

        //
        // Create a record
        //
        int key;
        {
            factory.DBRecord record = factory.createRecord();
            key = record.key();
            System.out.println("Created record: " + key);
            record.data("This is some data");
            String data = record.data();
            System.out.println("Added data: " + data);
        }

        //
        // Verify that getByKey works
        //
        try
        {
            System.out.println("Retrieving record again.");
            factory.DBRecord record = factory.getByKey(key);
            System.out.println("Got record: " + record.key());
            String data = record.data();
            System.out.println("Read data: " + data);
        }
        catch(factory.DBFactoryPackage.NoRecord ex)
        {
            ex.printStackTrace();
            System.exit(1);
        }

        //
        // Destroy the record
        //
        {
            factory.DBRecord record = null;
            try
            {
                record = factory.getByKey(key);
                System.out.println("Destroying record.");
                factory.destroyByKey(key);
            }
            catch(factory.DBFactoryPackage.NoRecord ex)
            {
                ex.printStackTrace();
                System.exit(1);
            }

            System.out.print("Expecting an OBJECT_NOT_EXIST exception...");
            System.out.flush();
            try
            {
                record.key();
                System.out.println("FAILED!");
            }
            catch(org.omg.CORBA.OBJECT_NOT_EXIST ex)
            {
                System.out.println("OK!");
            }

            System.out.print("Verifying that the record is destroyed...");
            System.out.flush();
            try
            {
                factory.DBRecord checkRecord = factory.getByKey(key);
                System.out.println("FAILED!");
            }
            catch(factory.DBFactoryPackage.NoRecord ex)
            {
                System.out.println("OK!");
            }
        }
    }

    //
    // Test a factory that uses a reaper
    //
    static void
    TestReaper(factory.DBFactory factory)
    {
        int key;

        //
        // Test that reaper works. This assumes that the timeout
        // and reaping periods on the server are extremely fast.
        //
        {
            System.out.println("Creating another record.");
            factory.DBRecord record = factory.createRecord();
            key = record.key();
            System.out.println("Created record: " + key);
            record.data("This is some more data");
            String data = record.data();
            System.out.println("Added data: " + data);

            System.out.println("Sleeping to wait for the record to be " +
                               "reaped.");
            try
            {
                Thread.sleep(1000 * 5);
            }
            catch(InterruptedException ex)
            {
            }

            System.out.println("Refetching object");
            try
            {
                record = factory.getByKey(key);

                data = record.data();
                System.out.println("Data is: " + data);

                factory.destroyByKey(key);
            }
            catch(factory.DBFactoryPackage.NoRecord ex)
            {
                ex.printStackTrace();
                System.exit(1);
            }
        }

        //
        // Create 10 records, then sleep to allow them all to be reaped
        //
        {
            int i;

            System.out.println("Creating 10 records");
            for(i = 0 ; i < 10 ; i++)
            {
                factory.DBRecord rec = factory.createRecord();
            }

            int[] records = factory.getAllKeys();

            System.out.println("" + records.length +
                               " records now in database.");

            //
            // We want all 10 records as  types
            //
            factory.DBRecord[] recs = new factory.DBRecord[10];
            try
            {
                for(i = 0 ; i < 10 ; i++)
                    recs[i] = factory.getByKey(records[i]);
            }
            catch(factory.DBFactoryPackage.NoRecord ex)
            {
                ex.printStackTrace();
                System.exit(1);
            }

            System.out.println("Sleeping five seconds.");
            try
            {
                Thread.sleep(1000 * 5);
            }
            catch(InterruptedException ex)
            {
            }
            //
            // The records should have been reaped, so we access them
            // again to force them to be reactivated.
            //
            System.out.println("Setting data");
            i = 0;
            while(i < 10)
            {
                String buf = "This is record: " + i;
                recs[i].data(buf);
                i++;
            }

            System.out.println("Sleeping five seconds.");
            try
            {
                Thread.sleep(1000 * 5);
            }
            catch(InterruptedException ex)
            {
            }
            System.out.println("Getting data");
            i = 0;
            while(i < 10)
            {
                String data = recs[i].data();
                System.out.println("Data: " + data);
                i++;
            }

            //
            // Purge these records
            //
            System.out.println("Purging " + records.length + " records.");
            for(i = 0 ; i < records.length ; i++)
            {
                try
                {
                    factory.destroyByKey(records[i]);
                }
                catch(factory.DBFactoryPackage.NoRecord ex)
                {
                    System.out.println("Record " + records[i] +
                                       " doesn't exist anymore.");
                }
            }
        }
    }

    static int
    run(org.omg.CORBA.ORB orb, String[] args)
        throws org.omg.CORBA.UserException
    {
        {
            //
            // Get factory object that uses a default servant
            //
            org.omg.CORBA.Object obj =
                orb.string_to_object("relfile:/dfactory.ref");
            if(obj == null)
            {
                System.err.println("cannot read IOR from dfactory.ref");
                return 1;
            }

            DBFactory factory = DBFactoryHelper.narrow(obj);

            System.out.println("----------------------------------------");
            System.out.println("Testing factory w/ default servant");
            System.out.println("----------------------------------------");
            TestFactory(factory);
        }

        {
            //
            // Get factory object that uses a servant locator
            //
            org.omg.CORBA.Object obj =
                orb.string_to_object("relfile:/lfactory.ref");
            if(obj == null)
            {
                System.err.println("factory.Client: " +
                                   "cannot read IOR from lfactory.ref");
                return 1;
            }

            DBFactory factory = DBFactoryHelper.narrow(obj);

            System.out.println();
            System.out.println("----------------------------------------");
            System.out.println("Testing factory w/ servant locator");
            System.out.println("----------------------------------------");
            TestFactory(factory);
            TestReaper(factory);
        }

        return 0;
    }

    public static void
    main(String[] args)
    {
        java.util.Properties props = System.getProperties();
        props.put("org.omg.CORBA.ORBClass", "com.ooc.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass",
                  "com.ooc.CORBA.ORBSingleton");

        int status = 0;
        org.omg.CORBA.ORB orb = null;

        try
        {
            orb = org.omg.CORBA.ORB.init(args, props);
            status = run(orb, args);
        }
        catch(Exception ex)
        {
            ex.printStackTrace();
            status = 1;
        }

        if(orb != null)
        {
            try
            {
                orb.destroy();
            }
            catch(Exception ex)
            {
                ex.printStackTrace();
                status = 1;
            }
        }

        System.exit(status);
    }
}
