// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package factory;

//
// Concrete subclass of Reaper that implements a least-recently-used
// algorithm for determining when to reap a servant
//
public class LRUReaper implements Reaper
{
    private static class ServantInfo
    {
        ServantInfo(org.omg.PortableServer.Servant s)
        {
            servant = s;
            count = 0;
            last = System.currentTimeMillis();
        }

        org.omg.PortableServer.Servant servant;
        int count;
        long last;
    }

    private static class ObjectIdHasher
    {
        private byte[] id_;
        private int hashCode_;

        ObjectIdHasher(byte[] id)
        {
            id_ = id;

            hashCode_ = 0;
            int offset = 0;
            int len = id.length;

            if(len < 16)
            {
                for(int i = len ; i > 0 ; i--)
                    hashCode_ = (hashCode_ * 37) + id[offset++];
            }
            else
            {
                int skip = len / 8;
                for(int i = len ; i > 0 ; i -= skip, offset += skip)
                    hashCode_ = (hashCode_ * 39) + id[offset];
            }
        }

        byte[]
        id()
        {
            return id_;
        }

	public int
        hashCode()
        {
            return hashCode_;
        }

	public boolean
        equals(java.lang.Object obj)
        {
            ObjectIdHasher h = (ObjectIdHasher)obj;
            if(id_.length != h.id_.length)
                return false;

            for(int i = 0 ; i < id_.length ; i++)
                if(id_[i] != h.id_[i])
                    return false;

            return true;
        }
    }

    private int maxIdle_;
    private ReapListener listener_;
    private java.util.Hashtable servantMap_;
    private boolean debug_;

    //
    // maxIdle defines the number of seconds a servant can be idle
    // until it becomes eligible for reaping. The listener, if
    // non-null, will be invoked when a servant is reaped.
    //
    public
    LRUReaper(boolean debug, int maxIdle, ReapListener listener)
    {
        maxIdle_ = maxIdle;
        listener_ = listener;
	debug_ = debug;
        servantMap_ = new java.util.Hashtable(1023);
    }

    //
    // Registers a servant with the reaper. If the reaper is currently
    // managing a servant with the same oid, nothing is done. Otherwise,
    // the reaper invokes _add_ref() once on the servant and begins
    // management of the servant.
    //
    public synchronized void
    addServant(byte[] oid, org.omg.PortableServer.Servant servant)
    {
        ObjectIdHasher key = new ObjectIdHasher(oid);
        if(!servantMap_.containsKey(key))
        {
            ServantInfo info = new ServantInfo(servant);
            servantMap_.put(key, info);
        }
    }

    //
    // Removes a servant. The reaper invokes _remove_ref() once on
    // the servant. The ReapListener (if any) is not invoked.
    //
    public synchronized void
    removeServant(byte[] oid)
    {
        ObjectIdHasher key = new ObjectIdHasher(oid);
        ServantInfo info = (ServantInfo)servantMap_.get(key);
        if(info != null)
            servantMap_.remove(key);
    }

    //
    // Obtains the servant with the given oid and signals that the
    // servant is "in use" by the application. If found, the servant
    // is returned, otherwise 0 is returned. The servant will not be
    // eligible for reaping until releaseServant() is called an
    // equivalent number of times.
    //
    public synchronized org.omg.PortableServer.Servant
    acquireServant(byte[] oid)
    {
        ObjectIdHasher key = new ObjectIdHasher(oid);
        ServantInfo info = (ServantInfo)servantMap_.get(key);

        if(info == null)
            return null;

        info.last = System.currentTimeMillis();
        info.count++;

        if(debug_)
        {
            String id = new String(oid);
            System.out.println("Touching servant " + id + ", count = " +
			       info.count);
        }

        return info.servant;
    }

    //
    // Indicates that the application is no longer using the servant
    // with the given oid. Each call to acquireServant() should
    // eventually be followed by a call to releaseServant().
    //
    public synchronized void
    releaseServant(byte[] oid)
    {
        ObjectIdHasher key = new ObjectIdHasher(oid);
        ServantInfo info = (ServantInfo)servantMap_.get(key);

        if(info == null)
            return;

        if(info.count <= 0)
            throw new RuntimeException();
        info.count--;

	if(debug_)
        {
            String id = new String(oid);
	    System.out.println("Releasing servant " + id + ", count = " +
			       info.count);
        }
    }

    //
    // Activates an immediate scan of the servants to reap any that
    // are eligible.
    //
    public synchronized void
    reap()
    {
	if(debug_)
	    System.out.println("Starting rep");

        long now = System.currentTimeMillis();

        java.util.Enumeration keys = servantMap_.keys();
        while(keys.hasMoreElements())
        {
            ObjectIdHasher key = (ObjectIdHasher)keys.nextElement();
            ServantInfo info = (ServantInfo)servantMap_.get(key);
            if(info == null)
                throw new RuntimeException();

            //
            // Determine the servant's age since last use
            //
            long age = (now - info.last) / 1000;

            //
            // Only reap the servant if the age exceeds maxIdle and the
            // usage count == 0 (the count will be 0 when the servant is
            // not currently in use by the application).
            //
            if(age > maxIdle_ && info.count == 0)
            {
		if(debug_)
                {
                    String id = new String(key.id());
                    System.out.println("servant " + id + ", age = " + age);
                }

                servantMap_.remove(key);
                if(listener_ != null)
                    listener_.servantReaped(key.id(), info.servant);
            }
        }
    }
}
