// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OB;

final class GIOPClient extends Client
{
    protected ORBInstance orbInstance_; // The ORB instance
    protected int nextRequestId_; // The next request ID
    protected java.lang.Object nextRequestIdMutex_ =
        new java.lang.Object(); // The next request ID mutex
    protected com.ooc.OCI.Connector connector_; // The connector
    protected GIOPClientWorker worker_; // The worker
    protected org.omg.IOP.ServiceContext codeSetSC_; // The codeset SC
    protected boolean destroy_; // True if destroy() was called

    // ----------------------------------------------------------------------
    // GIOPClient private and protected member implementations
    // ----------------------------------------------------------------------

    protected void
    finalize()
        throws Throwable
    {
        Assert._OB_assert(destroy_);
        Assert._OB_assert(worker_ == null);

        super.finalize();
    }

    //
    // Get the worker. If the bool flag is true, and no worker exists,
    // a new worker is created, with the timeout specified as second
    // parameter.
    //
    protected synchronized GIOPClientWorker
    getWorker(boolean create, int t)
    {
        if(destroy_)
            throw new org.omg.CORBA.INITIALIZE(
                MinorCodes.describeInitialize(MinorCodes.MinorORBDestroyed),
                MinorCodes.MinorORBDestroyed,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);

        if(worker_ == null && create)
        {
            //
            // Trace connection attempt
            //
            CoreTraceLevels coreTraceLevels =
                orbInstance_.getCoreTraceLevels();
            if(coreTraceLevels.traceConnections() > 0)
            {
                com.ooc.OCI.ConnectorInfo info = connector_.get_info();
                String msg = "trying to establish connection\n";
                msg += "timeout: ";
                if(t >= 0)
                {
                    msg += t;
                    msg += "ms\n";
                }
                else
                    msg += "none\n";
                msg += info.describe();
                orbInstance_.getLogger().trace("outgoing", msg);
            }

            //
            // Create new transport, using the connector
            //
            // For symetry reasons, GIOPClientStarterThreaded should also be
            // added, even though these classes only have a trivial
            // functionality. Or perhaps the GIOPClientStarterThreaded tries to
            // connect() in the backgound? Just an idea...
            //

            com.ooc.OCI.Transport transport;

            if(t >= 0)
            {
                transport = connector_.connect_timeout(t);

                //
                // Was there a timeout?
                //
                if(transport == null)
                    throw new org.omg.CORBA.NO_RESPONSE(
                        "Connection timeout", 0,
                        org.omg.CORBA.CompletionStatus.COMPLETED_NO);
            }
            else
            {
                transport = connector_.connect();
                Assert._OB_assert(transport != null);
            }

            //
            // Create new worker
            //

            Assert._OB_assert(concModel_ == Threaded);
            worker_ = new GIOPClientWorkerThreaded(orbInstance_, this,
                                                   transport);
	    //
	    // Lazy initialization of codeSetSC_. We don't want to
	    // initialize codeSetSC_ in the constructor, in order to
	    // keep creation of GIOPClients, which might later on not
	    // be used, as light-weight as possible.
	    //
	    if(codeSetSC_ == null)
	    {
		//
		// Create CONV_FRAME::CodeSetContext
		//
		org.omg.CONV_FRAME.CodeSetContext ctx =
		    new org.omg.CONV_FRAME.CodeSetContext();
		CodeConverters conv = codeConverters();
		if(conv.outputCharConverter != null)
		    ctx.char_data = conv.outputCharConverter.getTo().rgy_value;
		else
		    ctx.char_data = CodeSetDatabase.ISOLATIN1;
		if(conv.outputWcharConverter != null)
		    ctx.wchar_data = conv.outputWcharConverter.getTo().
			rgy_value;
		else
		    ctx.wchar_data = orbInstance_.getNativeWcs();

		//
		// Create encapsulation for CONV_FRAME::CodeSetContext
		//
		com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer();
		com.ooc.CORBA.OutputStream outCSC =
		    new com.ooc.CORBA.OutputStream(buf);
		outCSC._OB_writeEndian();
		org.omg.CONV_FRAME.CodeSetContextHelper.write(outCSC, ctx);
		
		//
		// Create service context containing the
		// CONV_FRAME::CodeSetContext encapsulation
		//
		codeSetSC_ = new org.omg.IOP.ServiceContext();
		codeSetSC_.context_id = org.omg.IOP.CodeSets.value;
		int len = buf.length();
		byte[] data = buf.data();
		codeSetSC_.context_data = new byte[len];
		System.arraycopy(data, 0, codeSetSC_.context_data, 0, len);
	    }
        }

        return worker_;
    }

    // ----------------------------------------------------------------------
    // GIOPClient package member implementations
    // ----------------------------------------------------------------------

    GIOPClient(ORBInstance orbInstance, com.ooc.OCI.Connector connector,
               int concModel, CodeConverters conv)
    {
        super(concModel, conv);
        orbInstance_ = orbInstance;
        connector_ = connector;
        destroy_ = false;
    }

    // ----------------------------------------------------------------------
    // GIOPClient public member implementations
    // ----------------------------------------------------------------------

    //
    // Destroy the client
    //
    public void
    destroy()
    {
        GIOPClientWorker worker = null;

        synchronized(this)
        {
            //
            // Don't destroy twice
            //
            if(destroy_)
                return;

            //
            // Set the destroy flag
            //
            destroy_ = true;

            //
            // Use a copy of the worker, and destroy the worker outside
            // the synchronization, to avoid deadlocks
            //
            worker = worker_;
            worker_ = null;
        }

        //
        // If there is a worker, destroy it
        //
        if(worker != null)
            worker.destroy();
    }

    //
    // Remove the worker. This operation is called by GIOPClientWorker
    // and derived classes only.
    //
    public synchronized void
    removeWorker(GIOPClientWorker worker)
    {
        if(worker_ == worker)
            worker_ = null;
    }

    //
    // Get a new request ID
    //
    public int
    requestId()
    {
        synchronized(nextRequestIdMutex_)
        {
            return nextRequestId_++;
        }
    }

    //
    // Get all profiles that are usable with this client
    //
    public com.ooc.OCI.ProfileInfo[]
    getUsableProfiles(org.omg.IOP.IOR ior, org.omg.CORBA.Policy[] policies)
    {
        //
        // Get all profiles usable for the connector
        //
        com.ooc.OCI.ProfileInfo[] all =
            connector_.get_usable_profiles(ior, policies);

        //
        // Filter out profiles which would require a different code converter
        //
        java.util.Vector vec = new java.util.Vector();
        for(int i = 0 ; i < all.length ; i++)
        {
            CodeConverters conv = CodeSetUtil.getCodeConverters(
		orbInstance_, all[i]);
            if(codeConverters().equals(conv))
                vec.addElement(all[i]);
        }

        com.ooc.OCI.ProfileInfo[] result =
            new com.ooc.OCI.ProfileInfo[vec.size()];
        vec.copyInto(result);
        return result;
    }

    //
    // Get the OCI Connector info
    //
    public com.ooc.OCI.ConnectorInfo
    connectorInfo()
    {
        return connector_.get_info();
    }

    //
    // Get the OCI Transport info
    //
    public com.ooc.OCI.TransportInfo
    transportInfo()
    {
        //
        // Get the worker, but do not create a new one if there is none
        // available
        //
        GIOPClientWorker worker = getWorker(false, -1);

        if(worker == null)
            return null;

        com.ooc.OCI.Transport transport = worker.transport();
        return transport.get_info();
    }

    //
    // Start a downcall, returning a downcall emitter and an
    // OutputStream for marshalling a request
    //
    public DowncallEmitter
    startDowncall(Downcall down, com.ooc.CORBA.OutputStreamHolder out)
    {
        GIOPClientWorker worker = null;
        try
        {
            //
            // Get the worker, creating a new one if there is none
            // available
            //
            worker = getWorker(true, down.policies().connectTimeout);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            Assert._OB_assert(ex.completed ==
                          org.omg.CORBA.CompletionStatus.COMPLETED_NO);
            down.setFailureException(ex);
            return null;
        }

        try
        {
            //
            // We only need to add a code set context if we're GIOP
            // version 1.1 or higher, and if no messages have been sent so
            // far
            //
            byte major = down.profileInfo().major;
            byte minor = down.profileInfo().minor;
            if(!worker.messageSent() && (major > 1 || minor >= 1))
            {
                CoreTraceLevels coreTraceLevels =
                    orbInstance_.getCoreTraceLevels();
                if(coreTraceLevels.traceConnections() >= 2)
                {
                    CodeConverters conv = codeConverters();
                    String msg = "sending transmission code sets";
                    msg += "\nchar code set: ";
		    if(conv.outputCharConverter != null)
			msg += conv.outputCharConverter.getTo().description;
		    else
		    {
			CodeSetInfo info =
			    CodeSetDatabase.instance().getCodeSetInfo(
				orbInstance_.getNativeCs());
			msg += info.description;
		    }
                    msg += "\nwchar code set: ";
                    if(conv.outputWcharConverter != null)
                        msg += conv.outputWcharConverter.getTo().description;
		    else
		    {
			CodeSetInfo info =
			    CodeSetDatabase.instance().getCodeSetInfo(
				orbInstance_.getNativeWcs());
			msg += info.description;
		    }
                    orbInstance_.getLogger().trace("outgoing", msg);
                }

		Assert._OB_assert(codeSetSC_ != null);
		down.addToRequestSCL(codeSetSC_);
	    }
	    
	    com.ooc.OCI.ProfileInfo profileInfo = down.profileInfo();
            com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer(12);
            buf.pos(12);
	    out.value =
		new com.ooc.CORBA.OutputStream(buf, codeConverters(),
					       (profileInfo.major << 8) |
					       profileInfo.minor);
	    
	    //
	    // Create GIOP outgoing message
	    //
            GIOPOutgoingMessage outgoing =
                new GIOPOutgoingMessage(orbInstance_, out.value, profileInfo);
	    
            //
            // Write header
            //
	    String op = down.operation();
	    if(op.charAt(0) == '_' && op.equals("_locate"))
	        outgoing.writeLocateRequestHeader(down.requestId());
	    else
                outgoing.writeRequestHeader(down.requestId(),
                                            down.operation(),
                                            down.responseExpected(),
                                            down.getRequestSCL());

            return worker;
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            Assert._OB_assert(ex.completed ==
                          org.omg.CORBA.CompletionStatus.COMPLETED_NO);
            down.setFailureException(ex);
            return null;
        }
    }

    //
    // Checks whether this client is equal to another client
    //
    public boolean
    equal(Client cl)
    {
        GIOPClient client = null;
        try
        {
            client = (GIOPClient)cl;
        }
        catch(ClassCastException ex)
        {
            return false;
        }

        if(!connector_.equal(client.connector_))
            return false;

        if(!codeConverters().equals(client.codeConverters()))
            return false;

        return true;
    }

    //
    // Force connection establishment
    //
    public void
    bind(int connectTimeout)
    {
        //
        // Get the worker, creating a new one if there is none
        // available
        //
        getWorker(true, connectTimeout);
    }

    //
    // Determines whether this client supports twoway invocations
    //
    public boolean
    twoway()
    {
        //
        // Get the worker
        //
        GIOPClientWorker worker = getWorker(false, -1);
        Assert._OB_assert(worker != null);
        com.ooc.OCI.Transport transport = worker.transport();
        return transport.mode() == com.ooc.OCI.SendReceiveMode.SendReceive;
    }
}
