// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OB;

abstract class GIOPClientWorker implements DowncallEmitter
{
    protected ORBInstance orbInstance_; // The ORBInstance
    protected GIOPClient client_; // The client
    protected com.ooc.OCI.Transport transport_; // The transport
    protected com.ooc.OCI.Buffer buf_; // The incoming message buffer
    protected GIOPIncomingMessage incoming_; // The incoming message
    protected boolean messageSent_; // True, if at least one message was sent
    protected byte minor_; // The highest minor version seen in all messages
    protected boolean closeLogged_; // True, if connection closure was already
                                    // logged
    protected int shutdownTimeout_; // The shutdown timeout

    //
    // Unsent messages
    //
    protected class UnsentMessage
    {
        com.ooc.OCI.Buffer buf;
        Downcall down;

        UnsentMessage(com.ooc.OCI.Buffer buf)
        {
            this.buf = buf;
            this.buf.pos(0);
        }

        UnsentMessage(Downcall down)
        {
            this.down = down;
            this.buf = down.output()._OB_buffer();
            this.buf.pos(0);
        }
    }
    protected java.util.Vector unsent_ = new java.util.Vector();

    //
    // Downcalls awaiting replies
    //
    protected java.util.Vector pending_ = new java.util.Vector();

    //
    // The worker's state
    //
    public final static int StateActive  = 0;
    public final static int StateClosing = 1;
    public final static int StateError   = 2;
    public final static int StateClosed  = 4;

    protected int state_;

    //
    // The failure exception. This must be set if state_ >=
    // StateError.
    //
    protected org.omg.CORBA.SystemException exception_;

    // ----------------------------------------------------------------------
    // GIOPClientWorker private and protected member implementations
    // ----------------------------------------------------------------------

    //
    // Emit a trace message when closing the connection
    //
    protected void
    logCloseConnection()
    {
        if(!closeLogged_)
        {
            CoreTraceLevels coreTraceLevels =
                orbInstance_.getCoreTraceLevels();
            if(coreTraceLevels.traceConnections() > 0)
            {
                com.ooc.OCI.TransportInfo info = transport_.get_info();
                String msg = "closing connection\n";
                msg += info.describe();
//                msg += '\n';
//                msg += downcallManager_.getFailureException().exception;
                orbInstance_.getLogger().trace("outgoing", msg);
            }

            closeLogged_ = true;
        }
    }

    //
    // Add message header
    //
    protected void
    addMessageHeader(Downcall down)
    {
        //
        // Save stream position and rewind
        //
        com.ooc.CORBA.OutputStream out = down.output();
        int pos = out._OB_pos();
        out._OB_pos(0);

        //
        // Write message header
        //
        com.ooc.OCI.ProfileInfo profileInfo = down.profileInfo();
        GIOPOutgoingMessage outgoing =
            new GIOPOutgoingMessage(orbInstance_, out, profileInfo);
        outgoing.writeMessageHeader(org.omg.GIOP.MsgType_1_1.Request, false,
                                    pos - 12);

        //
        // Restore stream position
        //
        out._OB_pos(pos);
    }

    //
    // Add new unsent buffer
    //

    protected void
    add(com.ooc.OCI.Buffer buf)
    {
        //
        // If we are in StateError or StateClosed, don't add.
        //
        if(state_ == StateError || state_ == StateClosed)
            return;
        
        //
        // Add new message to the message buffers
        //
        unsent_.addElement(new UnsentMessage(buf));
    }

    //
    // Add new unsent downcall
    //
    protected void
    add(Downcall down)
    {
        //
        // If a failure exception is set, don't add new downcall
        // objects. Just set the exception in the downcall.
        //
        if(exception_ != null)
        {
            down.setFailureException(exception_);
            return;
        }

        //
        // If we are in StateClosing, StateError, or StateClosed, then why
        // is no exception set?
        //
        Assert._OB_assert(state_ != StateClosing && state_ != StateError &&
                      state_ != StateClosed);

        //
        // Add the message header
        //
        try
        {
            //
            // Save stream position and rewind
            //
            com.ooc.CORBA.OutputStream out = down.output();
            int pos = out._OB_pos();
            out._OB_pos(0);

            //
            // Write message header
            //
            com.ooc.OCI.ProfileInfo profileInfo = down.profileInfo();
            GIOPOutgoingMessage outgoing =
                new GIOPOutgoingMessage(orbInstance_, out, profileInfo);

            //
            // Write protocol header
            //
            String op = down.operation();
            if(op.charAt(0) == '_' && op.equals("_locate"))
	        outgoing.writeMessageHeader(
	            org.omg.GIOP.MsgType_1_1.LocateRequest, false, pos - 12);
            else
	        outgoing.writeMessageHeader(
	            org.omg.GIOP.MsgType_1_1.Request, false, pos - 12);

            //
            // Restore stream position
            //
            out._OB_pos(pos);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            Assert._OB_assert(ex.completed ==
                          org.omg.CORBA.CompletionStatus.COMPLETED_NO);
            down.setFailureException(ex);
            return;
        }

        //
        // Update the GIOP minor version
        //
        byte minor = down.profileInfo().minor;
        if(minor > minor_)
            minor_ = minor;

        //
        // Add new message to the message buffers
        //
        unsent_.addElement(new UnsentMessage(down));
    }

    //
    // Move the first unsent downcall to the list of pending downcalls
    //
    protected Downcall
    moveFirstUnsentToPending()
    {
        //
        // If there are no unsent messages, then there must be a failure
        // exception, otherwise it's a bug.
        //
        if(unsent_.isEmpty())
        {
            Assert._OB_assert(exception_ != null);
            return null;
        }

        //
        // Remove first downcall
        //
        UnsentMessage m = (UnsentMessage)unsent_.firstElement();
        Downcall down = m.down;
        unsent_.removeElementAt(0);

        //
        // Only add to pending if a response is expected
        //
        if(down != null)
        {
            if(down.responseExpected())
            {
                down.setPending();
                pending_.addElement(down);
            }
            else
            {
                down.setNoException(null);
            }
        }

	//
	// Do not set messageSent_ if operation was a LocateRequest since
	// LocateRequest does not contain ServiceContext information
	//
	if(!messageSent_)
	{
            if(!down.operation().equals("_locate"))
	    {
	        messageSent_ = true;
	    }
	}

	return down;
    }

    //
    // Remove pending message and put it back in the unsent list
    //
    protected void
    movePendingToUnsent(Downcall down)
    {
	if(down.responseExpected())
	{
	    Downcall d = findAndRemovePending(down.requestId());
	    Assert._OB_assert(d == null || d == down);
	    unsent_.addElement(new UnsentMessage(down));
	}
    }

    //
    // Find and remove a pending downcall
    //
    protected Downcall
    findAndRemovePending(int reqId)
    {
        for(int i = 0 ; i < pending_.size() ; i++)
        {
            Downcall d = (Downcall)pending_.elementAt(i);
            if(d.requestId() == reqId)
            {
                pending_.removeElementAt(i);
                return d;
            }
        }

        return null;
    }

    //
    // Change the state of the worker due to an exception
    //
    protected boolean
    exception(int state, org.omg.CORBA.SystemException ex,
              boolean notCompleted)
    {
        //
        // Don't do anything if there is no state change
        //
        if(state_ == state)
            return false;

        //
        // It is not possible to transition backwards
        //
        if(state < state_)
            return false;

        //
        // Set the failure exception
        //
        if(exception_ == null)
        {
            //
            // Note: The completion status of ex is not taken into account
            // here, since the completion status of an exception raised by
            // low-level operations (like OCI send and receive methods) is
            // not applicable as the completion status of the downcall as
            // a whole. Therefore the completion status is passed
            // explicitly with the notCompleted parameter.
            //

            //
            // Save the exception, with a completion status of COMPLETED_NO
            //
            exception_ = Util.copySystemException(ex);
            exception_.completed = org.omg.CORBA.CompletionStatus.COMPLETED_NO;

            //
            // Always use a completion status of NO for unsent requests
            //
            while(!unsent_.isEmpty())
            {
                UnsentMessage m = (UnsentMessage)unsent_.firstElement();
                if(m.down != null)
                    m.down.setFailureException(exception_);

                unsent_.removeElementAt(0);
            }

            //
            // If notCompleted is true, use a completion status of
            // COMPLETED_NO for pending requests, otherwise use
            // COMPLETED_MAYBE
            //
            org.omg.CORBA.SystemException except =
                Util.copySystemException(exception_);
            if(!notCompleted)
                except.completed =
                    org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE;
            while(!pending_.isEmpty())
            {
                Downcall down = (Downcall)pending_.firstElement();
                down.setFailureException(except);
                pending_.removeElementAt(0);
            }
        }

        switch(state)
        {
        case StateActive:

            //
            // There is nothing special to do
            //
            break;

        case StateClosing:

            //
            // Detach the worker from the client
            //
            client_.removeWorker(this);

            // 
            // Do a graceful shutdown after state change below...
            // 

            break;

        case StateError:

            //
            // Detach the worker from the client
            //
            client_.removeWorker(this);

            // 
            // Do an abortive shutdown
            // 
            abortiveShutdown();
            
            break;

        case StateClosed:

            //
            // Detach the worker from the client
            //
            client_.removeWorker(this);

            //
            // Close the transport
            //
            logCloseConnection();
            transport_.close();

            break;
        }

        //
        // Update the state
        //
        state_ = state;

	//
	// If StateClosing, shutdown gracefully. This must be called after
	// the state has been changed since calling gracefulShutdown() may
	// cause a state change to StateClosed and we do not want to reset it
	// to StateClosing if that occurs.
	//
	if(state_ == StateClosing)
	    gracefulShutdown();

        return true;
    }

    //
    // For consistency with C++
    //
    protected final void
    exception(int state, org.omg.CORBA.SystemException ex)
    {
        exception(state, ex, false);
    }

    // 
    // Shutdown the connection gracefully
    //
    protected abstract void
    gracefulShutdown();

    // 
    // Abortive connection shutdown
    //
    protected abstract void
    abortiveShutdown();
    
    //
    // Execute incoming message
    //
    protected void
    execute()
    {
        Assert._OB_assert(state_ == StateActive || state_ == StateClosing);

        boolean complete;
        try
        {
            //
            // Provide the buffer to the incoming message and reset buf_
            // to allow for reading the next message. A return value
            // of false indicates that a message is not available (i.e., a
            // fragment was received).
            //
            com.ooc.OCI.Buffer buf = buf_;
            buf_ = null;
            complete = incoming_.consumeBuffer(buf);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            exception(StateError, ex);
            return;
        }

        //
        // Keep track of the highest GIOP minor version seen
        //
        byte minor = incoming_.version().minor;
        if(minor > minor_)
            minor_ = minor;

        //
        // If the message is not complete yet, there is nothing to do
        //
        if(!complete)
            return;

        //
        // OK, message is complete - let's execute it
        //
        switch(incoming_.type().value())
        {
        case org.omg.GIOP.MsgType_1_1._Reply:
        {
	    int reqId = 0;
	    org.omg.GIOP.ReplyStatusType_1_2Holder status =
		new org.omg.GIOP.ReplyStatusType_1_2Holder();
	    org.omg.IOP.ServiceContextListHolder scl =
		new org.omg.IOP.ServiceContextListHolder();

	    try
	    {
                reqId = incoming_.readReplyHeader(status, scl);
	    }
            catch(org.omg.CORBA.SystemException ex)
            {
                exception(StateError, ex);
                return;
            }

            Downcall down = findAndRemovePending(reqId);
            if(down == null)
            {
                //
                // Request id is unknown
                //
                exception(StateError,
                          new org.omg.CORBA.COMM_FAILURE(
                              MinorCodes.describeCommFailure(
                                  MinorCodes.MinorUnknownReqId) + ": " + reqId,
                              MinorCodes.MinorUnknownReqId,
                              org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
                return;
            }

            down.setReplySCL(scl.value);
            com.ooc.CORBA.InputStream in = incoming_.input();

            switch(status.value.value())
            {
            case org.omg.GIOP.ReplyStatusType_1_2._NO_EXCEPTION:
                down.setNoException(in);
                break;

            case org.omg.GIOP.ReplyStatusType_1_2._USER_EXCEPTION:
                down.setUserException(in);
                break;

            case org.omg.GIOP.ReplyStatusType_1_2._SYSTEM_EXCEPTION:
            {
                try
                {
                    org.omg.CORBA.SystemException ex =
                        Util.unmarshalSystemException(in);
                    //                    ex.remote(true);
                    down.setSystemException(ex);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    exception(StateError, ex);
                }

                break;
            }

            case org.omg.GIOP.ReplyStatusType_1_2._LOCATION_FORWARD:
            {
                try
                {
                    org.omg.IOP.IOR ior = org.omg.IOP.IORHelper.read(in);
                    down.setLocationForward(ior, false);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    exception(StateError, ex);
                }
                
                break;
            }

            case org.omg.GIOP.ReplyStatusType_1_2._LOCATION_FORWARD_PERM:
            {
                try
                {
                    org.omg.IOP.IOR ior = org.omg.IOP.IORHelper.read(in);
                    down.setLocationForward(ior, true);
                    break;
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    exception(StateError, ex);
                }
            }

            case org.omg.GIOP.ReplyStatusType_1_2._NEEDS_ADDRESSING_MODE:
                // TODO: implement
                exception(StateError,
                          new org.omg.CORBA.NO_IMPLEMENT(
                              MinorCodes.describeNoImplement(
                                  MinorCodes.MinorNotSupportedByLocalObject),
                              MinorCodes.MinorNotSupportedByLocalObject,
                              org.omg.CORBA.CompletionStatus.COMPLETED_NO));
                break;

            default:
                exception(StateError,
                          new org.omg.CORBA.COMM_FAILURE(
                              MinorCodes.describeCommFailure(
                                  MinorCodes.MinorUnknownReplyMessage),
                              MinorCodes.MinorUnknownReplyMessage,
                              org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
                break;
            }

            break;
        }

        case org.omg.GIOP.MsgType_1_1._Request:
        case org.omg.GIOP.MsgType_1_1._LocateRequest:
        case org.omg.GIOP.MsgType_1_1._CancelRequest:

            //
            // These messages may not be sent to a client
            //
            exception(StateError,
                      new org.omg.CORBA.COMM_FAILURE(
                          MinorCodes.describeCommFailure(
                              MinorCodes.MinorWrongMessage),
                          MinorCodes.MinorWrongMessage,
                          org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
            break;

        case org.omg.GIOP.MsgType_1_1._LocateReply:
        {
            int reqId;
            org.omg.GIOP.LocateStatusType_1_2Holder status =
	        new org.omg.GIOP.LocateStatusType_1_2Holder();

            try
            {
	        reqId = incoming_.readLocateReplyHeader(status);
            }
            catch(org.omg.CORBA.SystemException ex)
            {
	        exception(StateError, ex);
	        return;
            }

            Downcall down = findAndRemovePending(reqId);
            if(down == null)
            {
	        //
	        // Request id is unknown
	        //
	        exception(StateError,
		          new org.omg.CORBA.COMM_FAILURE(
                              MinorCodes.describeCommFailure(
                                  MinorCodes.MinorUnknownReqId),
		              MinorCodes.MinorUnknownReqId,
		              org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
	        return;
            }

            //
            // Was this a LocateRequest?
            //
            String op = down.operation();
            if(!op.equals("_locate"))
            {
	        exception(StateError,
		          new org.omg.CORBA.COMM_FAILURE(
                              MinorCodes.describeCommFailure(
                                  MinorCodes.MinorWrongMessage),
		              MinorCodes.MinorWrongMessage,
		              org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
	        return;
            }

	    com.ooc.CORBA.InputStream in = incoming_.input();

	    switch(status.value.value())
	    {
	    case org.omg.GIOP.LocateStatusType_1_2._UNKNOWN_OBJECT:
		down.setSystemException(new org.omg.CORBA.OBJECT_NOT_EXIST());
	        break;

	    case org.omg.GIOP.LocateStatusType_1_2._OBJECT_HERE:
		down.setNoException(in);
	        break;

	    case org.omg.GIOP.LocateStatusType_1_2._OBJECT_FORWARD:
                try
                {
                    org.omg.IOP.IOR ior = org.omg.IOP.IORHelper.read(in);
                    down.setLocationForward(ior, false);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    Logger logger = orbInstance_.getLogger();
                    logger.warning("An error occurred while reading a " +
                                   "locate reply, possibly indicating\n" +
                                   "an interoperability problem. You may " +
                                   "need to set the LocateRequestPolicy\n" +
                                   "to false.");
                    down.setSystemException(ex);
                    exception(StateError, ex);
                }

	        break;

	    case org.omg.GIOP.LocateStatusType_1_2._OBJECT_FORWARD_PERM:
                try
                {
                    org.omg.IOP.IOR ior = org.omg.IOP.IORHelper.read(in);
                    down.setLocationForward(ior, true);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    Logger logger = orbInstance_.getLogger();
                    logger.warning("An error occurred while reading a " +
                                   "locate reply, possibly indicating\n" +
                                   "an interoperability problem. You may " +
                                   "need to set the LocateRequestPolicy\n" +
                                   "to false.");
                    down.setSystemException(ex);
                    exception(StateError, ex);
                }

	        break;

	    case org.omg.GIOP.LocateStatusType_1_2._LOC_SYSTEM_EXCEPTION:
                try
                {
                    org.omg.CORBA.SystemException ex =
                        org.omg.CORBA.SystemExceptionHelper.read(in);
                    // ex.remote(true); // Not possible in Java
                    down.setSystemException(ex);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    down.setSystemException(ex);
                    exception(StateError, ex);
                }

	        break;

	    case org.omg.GIOP.LocateStatusType_1_2._LOC_NEEDS_ADDRESSING_MODE:
		// TODO: implement
		exception(StateError, new org.omg.CORBA.NO_IMPLEMENT());
	        break;
	    }

	    break;
	}

        case org.omg.GIOP.MsgType_1_1._CloseConnection:

            //
            // If the peer closes the connection, all outstanding requests
            // can safely be reissued. Thus we send all of them a
            // TRANSIENT exception with a completion status of
            // COMPLETED_NO. This is done by calling exception() with the
            // notCompleted parameter set to true.
            //
            exception(StateClosed,
                      new org.omg.CORBA.TRANSIENT(
                          MinorCodes.describeTransient(
                              MinorCodes.MinorCloseConnection),
                          MinorCodes.MinorCloseConnection,
                          org.omg.CORBA.CompletionStatus.COMPLETED_NO),
                      true);
            break;

        case org.omg.GIOP.MsgType_1_1._MessageError:

            //
            // How should we react to MessageError messages? We just send
            // a MessageError back, and close the connection.
            //
            exception(StateError,
                      new org.omg.CORBA.COMM_FAILURE(
                          MinorCodes.describeCommFailure(
                              MinorCodes.MinorMessageError),
                          MinorCodes.MinorMessageError,
                          org.omg.CORBA.CompletionStatus.COMPLETED_NO));
            break;

        case org.omg.GIOP.MsgType_1_1._Fragment:

            //
            // At this point there should be no fragments, only complete
            // messages
            //
            Assert._OB_assert(false);
            break;

        default:
            exception(StateError,
                      new org.omg.CORBA.COMM_FAILURE(
                          MinorCodes.describeCommFailure(
                              MinorCodes.MinorUnknownMessage),
                          MinorCodes.MinorUnknownMessage,
                          org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
            break;
        }
    }

    protected void
    finalize()
        throws Throwable
    {
        Assert._OB_assert(state_ == StateClosed);

        super.finalize();
    }

    // ----------------------------------------------------------------------
    // GIOPClientWorker package member implementations
    // ----------------------------------------------------------------------

    GIOPClientWorker(ORBInstance orbInstance, GIOPClient client,
                     com.ooc.OCI.Transport transport)
    {
        orbInstance_ = orbInstance;
        client_ = client;
	transport_ = transport;
        incoming_ = new GIOPIncomingMessage(orbInstance_);
        messageSent_ = false;
        minor_ = 0;
        closeLogged_ = false;
        state_ = StateActive;

        //
        // Cache properties
        //
        java.util.Properties properties = orbInstance_.getProperties();
        String value;

        value = properties.getProperty("ooc.orb.client_shutdown_timeout");
        if(value != null)
            shutdownTimeout_ = Integer.parseInt(value);
        else
            shutdownTimeout_ = 2;

        //
        // Trace new outgoing connection
        //
        CoreTraceLevels coreTraceLevels = orbInstance_.getCoreTraceLevels();
        if(coreTraceLevels.traceConnections() > 0)
        {
            com.ooc.OCI.TransportInfo info = transport_.get_info();
            String msg = "new connection\n";
            msg += info.describe();
            orbInstance_.getLogger().trace("outgoing", msg);
        }
    }

    //
    // Destroy the worker
    //
    abstract void
    destroy();

    //
    // Get the OCI transport
    //
    final com.ooc.OCI.Transport
    transport()
    {
        return transport_;
    }

    //
    // Return true if at least one message has been sent successfully
    //
    final boolean
    messageSent()
    {
        return messageSent_;
    }
}
