// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OB;

final class GIOPServerStarterThreaded extends GIOPServerStarter
{
    //
    // The starter thread
    //
    protected final class StarterThread extends Thread
    {
	private GIOPServerStarterThreaded starter_;
	
        StarterThread(ThreadGroup group, GIOPServerStarterThreaded starter)
        {
	    super(group, "ORBacus:Server:StarterThread");
	    starter_ = starter;
        }
        
	public void
        run()
	{
            try
            {
                starter_.starterRun();
            }
            catch(RuntimeException ex)
            {
                ex.printStackTrace();
                Assert._OB_assert(false);
            }

            //
            // Shutdown the acceptor so that no further connections are
            // accepted
            //
            starter_.logCloseAcceptor();
            starter_.acceptor_.shutdown();

            //
            // Accept all connections which might have queued up in the
            // listen() backlog
            //
            while(true)
            {
                com.ooc.OCI.Transport transport = null;

                try
                {
                    transport = starter_.acceptor_.accept(false);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                }

                if(transport == null)
                    break;

                try
                {
                    GIOPServerWorker worker =
                        new GIOPServerWorkerThreaded(starter_.orbInstance_,
                                                     transport,
                                                     starter_.oaInterface_);

                    worker.setState(GIOPServerWorker.StateClosing);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    // Ignore SystemExceptions
                }
            }

            //
            // Close the acceptor
            //
            starter_.acceptor_.close();

            //
            // Break cyclic object dependency
	    //
	    starter_ = null;
 	}
    }
    protected Thread starterThread_;

    // ----------------------------------------------------------------------
    // GIOPServerStarterThreaded package member implementation
    // ----------------------------------------------------------------------

    GIOPServerStarterThreaded(ORBInstance orbInstance,
                              com.ooc.OCI.Acceptor acceptor,
			      OAInterface oaInterface)
    {
	super(orbInstance, acceptor, oaInterface);

        try
        {
            //
            // Retrieve the thread group for the servers
            //
            ThreadGroup group = orbInstance_.getServerWorkerGroup();

            //
            // Start starter thread
            //
            starterThread_ = new StarterThread(group, this);
            starterThread_.start();
        }
        catch(OutOfMemoryError ex)
        {
            acceptor_.close();
            state_ = StateClosed;
            throw new org.omg.CORBA.IMP_LIMIT(
                MinorCodes.describeImpLimit(MinorCodes.MinorThreadLimit),
                MinorCodes.MinorThreadLimit,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    // ----------------------------------------------------------------------
    // GIOPServerStarterThreaded public member implementation
    // ----------------------------------------------------------------------

    //
    // Change the state of the worker
    //
    synchronized public void
    setState(int state)
    {
        //
        // Don't do anything if there is no state change
        //
        if(state_ == state)
            return;

        //
        // It is not possible to transition backwards, except if we are
        // in holding state
        //
        if(state_ != StateHolding && state < state_)
            return;

        switch(state)
        {
        case StateActive:
	{
	    for(int i = 0 ; i < workers_.size() ; i++)
	    {
	        GIOPServerWorker w = (GIOPServerWorker)workers_.elementAt(i);
	        w.setState(GIOPServerWorker.StateActive);
	    }
	    
	    break;
	}

        case StateHolding:
	    for(int i = 0 ; i < workers_.size() ; i++)
	    {
	        GIOPServerWorker w = (GIOPServerWorker)workers_.elementAt(i);
	        w.setState(GIOPServerWorker.StateHolding);
	    }
            break;

        case StateClosed:
	{
            for(int i = 0 ; i < workers_.size() ; i++)
            {
                GIOPServerWorker w = (GIOPServerWorker)workers_.elementAt(i);
                w.setState(GIOPServerWorker.StateClosing);
            }
            workers_.removeAllElements();

            //
            // Connect to this starter's acceptor, to unblock the call
            // to accept() in the starter thread
            //
            try
            {
                com.ooc.OCI.Transport tr = acceptor_.connect_self();
                tr.close();
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                //
                // Ignore all system exceptions
                //
            }

            break;
	}
        }

        //
        // Update the state and notify about the state change
        //
        state_ = state;
        notifyAll();
    }

    //
    // Run method for starter thread
    //
    public void
    starterRun()
    {
        while(true)
        {
            //
            // Get new transport, blocking
            //
            com.ooc.OCI.Transport transport = null;
            try
            {
                transport = acceptor_.accept(true);
                Assert._OB_assert(transport != null);
            }
            catch(org.omg.CORBA.NO_PERMISSION ex)
            {
                //
                // Ignore NO_PERMISSION exceptions
                //
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                //
                // Ignore exception. This probably means that the server
                // exceeded the number of available file descriptors.
                //
            }


            synchronized(this)
            {
	        //
		// Reap the existing set of workers
		//
		reapWorkers();

                //
                // Check whether we are on hold
                //
                while(state_ == StateHolding)
                {
                    try
                    {
                        wait();
                    }
                    catch(InterruptedException ex)
                    {
                    }
                }

                if(transport != null)
                {
                    try
                    {
                        if(state_ == StateActive)
                        {
                            //
                            // If we're active, we create and add a new
                            // worker to the worker list
                            //
                            GIOPServerWorker worker =
                                new GIOPServerWorkerThreaded(orbInstance_,
                                                             transport,
                                                             oaInterface_);
                            workers_.addElement(worker);
                            worker.setState(GIOPServerWorker.StateActive);
                        }
                        else
                        {
                            //
                            // If we're closed, we create a new dummy
                            // worker, only in order to set it to
                            // StateClosing for proper connection shutdown
                            //
                            Assert._OB_assert(state_ == StateClosed);

                            GIOPServerWorker worker =
                                new GIOPServerWorkerThreaded(orbInstance_,
                                                             transport,
                                                             oaInterface_);

                            worker.setState(GIOPServerWorker.StateClosing);
                        }
                    }
                    catch(org.omg.CORBA.SystemException ex)
                    {
                        String msg = "can't accept connection\n" +
                            ex.getMessage();
                        orbInstance_.getLogger().warning(msg);
                    }
                }

                if(state_ == StateClosed)
                    break;
            }
        }
    }
}
