// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OCI.UDP;

final class ServerPacketHandler extends Thread
{
    //
    // The acceptor info
    //
    private com.ooc.OCI.AcceptorInfo acceptorInfo_;

    //
    // The host
    //
    private String host_;

    //
    // The port
    //
    private int port_;

    //
    // The address
    //
    private java.net.InetAddress addr_;

    //
    // The socket
    //
    private java.net.DatagramSocket fd_;

    //
    // The trace level
    //
    private int trace_;

    //
    // The message timeout
    //
    private int messageTimeout_;

    //
    // The transport timeout
    //
    private int transportTimeout_;

    //
    // The logger
    //
    private com.ooc.OB.Logger logger_;

    //
    // The active transports
    //
    private java.util.Hashtable transports_ = new java.util.Hashtable();

    //
    // Transports yet to be accepted
    //
    private java.util.Vector newTransports_ = new java.util.Vector();
    private java.lang.Object newTransportsMonitor_ = new java.lang.Object();

    final static class TransportKey
    {
        int clientID;
        java.net.InetAddress remoteAddr;
        int remotePort;

        public boolean
        equals(Object obj)
        {
            TransportKey key = (TransportKey)obj;
            return clientID == key.clientID &&
                   remoteAddr.equals(key.remoteAddr) &&
                   remotePort == key.remotePort;
        }

        public int
        hashCode()
        {
            return clientID ^ remoteAddr.hashCode() ^ remotePort; 
        }
    }

    ServerPacketHandler(com.ooc.OCI.AcceptorInfo acceptorInfo,
                        String host, int port, java.net.InetAddress addr,
                        java.net.DatagramSocket fd,
                        int trace, int messageTimeout, int transportTimeout,
                        com.ooc.OB.Logger logger)
    {
        super("ORBacus:UDP:ServerPacketHandler");

        acceptorInfo_ = acceptorInfo;
        host_ = host;
        port_ = port;
        addr_ = addr;
        fd_ = fd;
        trace_ = trace;
        messageTimeout_ = messageTimeout;
        transportTimeout_ = transportTimeout;
        logger_ = logger;

        //
        // Set a timeout on the socket to workaround bugs on
        // some platforms where closing the socket doesn't unblock
        // a thread waiting in receive()
        //
        try
        {
            fd_.setSoTimeout(500); // 500 msec
        }
        catch(java.net.SocketException ex)
        {
            throw new org.omg.CORBA.COMM_FAILURE(
                com.ooc.OB.MinorCodes.describeCommFailure(
                    com.ooc.OB.MinorCodes.MinorSocket) +
                    ": " + ex.toString(),
                com.ooc.OB.MinorCodes.MinorSocket,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    public void
    run()
    {
        while(true)
        {
            //
            // Allocate a packet, large enough to hold all possible
            // datagram sizes
            //
            com.ooc.OCI.Buffer packet = new com.ooc.OCI.Buffer(65535);

            java.net.DatagramSocket fd = fd_; // avoid race condition
            java.net.DatagramPacket datagramPacket = null;
            if(fd != null)
            {
                //
                // Receive a packet
                //
                datagramPacket = new java.net.DatagramPacket(packet.data(),
                                                             packet.length());
                try
                {
                    fd.receive(datagramPacket);
                }
                catch(java.io.InterruptedIOException ex)
                {
                    //
                    // A timeout occurred - check if the socket was closed
                    //
                    if(fd_ == null)
                        break;

                    if(ex.bytesTransferred == 0)
                        continue;
                }
                catch(java.io.IOException ex)
                {
                    if(fd_ == null)
                        break;

                    //                
                    // TODO: What to do here ? Throwing an exception ? Nobody
                    // would catch it...
                    //
                    com.ooc.OB.Assert._OB_assert(false);
                    throw new org.omg.CORBA.COMM_FAILURE(
                        ex.toString(), com.ooc.OB.MinorCodes.MinorRecv,
                        org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE);
                }
            }

            if(fd_ == null)
                break;

            if(datagramPacket.getLength() <= 0)
            {
                //                
                // TODO: What to do here ? Throwing an exception ? Nobody
                // would catch it...
                //
                com.ooc.OB.Assert._OB_assert(false);
                throw new org.omg.CORBA.COMM_FAILURE(
                    "", com.ooc.OB.MinorCodes.MinorRecvZero, 
                    org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE);
            }

            //
            // Save the packet size
            //
            int packetSize = datagramPacket.getLength();

            //
            // Unmarshal the fixed header
            //
            com.ooc.CORBA.InputStream in = 
                new com.ooc.CORBA.InputStream(packet, 0, false);
            com.ooc.UDPIOP.FixedPacketHeader fHeader = 
                com.ooc.UDPIOP.FixedPacketHeaderHelper.read(in);

            // 
            // Now extract the endian-ness from the packet
            //
            boolean swap = ( ((fHeader.flags & 0x01) == 1) != false );

            //
            // Unmarshal the fixed header again, swapping bytes if "swap"
            // is true
            //
            if(swap)
            {
                in = new com.ooc.CORBA.InputStream(packet, 0, swap);
                fHeader = com.ooc.UDPIOP.FixedPacketHeaderHelper.read(in);
            }

            //
            // Construct a key from the remote address and port
            //
            TransportKey key = new TransportKey();
            key.clientID = fHeader.client_id;
            key.remoteAddr = datagramPacket.getAddress();
            key.remotePort = datagramPacket.getPort();

            while(true)
            {
                //
                // Try to find the transport that matches the key
                //
                ServerTransport_impl transport =
                    (ServerTransport_impl)transports_.get(key);

                if(transport == null)
                {
                    //
                    // No transport matches the key, so create a new transport
                    //
                    transport = new ServerTransport_impl(acceptorInfo_,
                                                         key.clientID,
                                                         key.remoteAddr,
                                                         key.remotePort,
                                                         this,
                                                         trace_,
                                                         messageTimeout_,
                                                         transportTimeout_,
                                                         logger_);

                    Object old = transports_.put(key, transport);

                    //
                    // Return value should be null since there no
                    // previous entry corresponding to this key
                    //
                    if(old != null)
                        throw new InternalError();

                    //
                    // Add the new transport to the new transport list, and
                    // notify threads waiting for new transports
                    //
                    synchronized(newTransportsMonitor_)
                    {
                        newTransports_.addElement(transport);
                        newTransportsMonitor_.notify();
                    }
                }

                //
                // Add the packet to the transport. If the return
                // value is true, the transport is active and has
                // accepted the packet. Otherwise we must create
                // a new transport.
                //
                if(transport.addPacket(in, packetSize, fHeader))
                    break;

                transports_.remove(key);
            }
        }
    }

    void
    removeTransport(int clientID, java.net.InetAddress remoteAddr,
                    int remotePort)
    {
        //
        // Construct a key from the client ID, remote address and port
        //
        TransportKey key = new TransportKey();
        key.clientID = clientID;
        key.remoteAddr = remoteAddr;
        key.remotePort = remotePort;

        //
        // Remove the transport that matches the key
        //
        transports_.remove(key);
    }

    com.ooc.OCI.Transport
    getNewTransport(boolean block)
    {
        synchronized(newTransportsMonitor_)
        {
            if(newTransports_.isEmpty())
            {
                if(!block)
                    return null;
                else
                {
                    while(newTransports_.isEmpty())
                    {
                        try
                        {
                            newTransportsMonitor_.wait();
                        }
                        catch(InterruptedException ex)
                        {
                        }
                    }
                }
            }

            com.ooc.OCI.Transport transport =
                (com.ooc.OCI.Transport)newTransports_.firstElement();
            newTransports_.removeElementAt(0);

            return transport;
        }
    }

    void
    connectSelf(com.ooc.OCI.Transport t)
    {
        synchronized(newTransportsMonitor_)
        {
            newTransports_.addElement(t);
            newTransportsMonitor_.notify();
        }
    }

    void
    close()
    {
        java.net.DatagramSocket savefd = fd_;
        fd_ = null;
        savefd.close();
    }

    void
    shutdown()
    {
        //
        // Not supported in Java
        //
    }

    String
    getHost()
    {
        return host_;
    }

    int
    getPort()
    {
        return port_;
    }

    java.net.InetAddress
    getAddress()
    {
        return addr_;
    }
}
