// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.LoadBalancing.Util;

import com.ooc.LoadBalancing.*;

//
// IDL:ooc.com/LoadBalancing/LoadUpdater:1.0
//
public class LoadUpdater_impl
    extends org.omg.CORBA.LocalObject
    implements LoadUpdater
{
    //
    // The LoadUpdater thread
    //
    private LoadUpdaterThread updater_;

    // -----------------------------------------------------------------
    // LoadUpdater_impl constructor
    // -----------------------------------------------------------------

    public LoadUpdater_impl(Logger logger, GroupFactory factory, int freq)
    {
	updater_ = new LoadUpdaterThread(logger, factory, freq);

	//
	// Start the updater thread.
	//
	updater_.start();
    }

    protected void finialize()
	throws Throwable
    {
	//
	// Stop the updater thread.
	//
	updater_.destroy();
	while(updater_.isAlive())
	{
	    try
	    {
		updater_.join();
	    }
	    catch(InterruptedException ex)
	    {
	    }
	}
    }

    // -----------------------------------------------------------------
    // IDL to Java Mapping
    // -----------------------------------------------------------------

    //
    // IDL:ooc.com/LoadBalancing/LoadUpdater/get_update_frequency:1.0
    //
    public int get_update_frequency()
    {
	return updater_.getFrequency();
    }

    //
    // IDL:ooc.com/LoadBalancing/LoadUpdater/set_update_frequency:1.0
    //
    public void set_update_frequency(int millis)
    {
	updater_.setFrequency(millis);
    }

    //
    // IDL:ooc.com/LoadBalancing/LoadUpdater/set_load_calculator:1.0
    //
    public void set_load_calculator(LoadCalculator calc)
    {
	updater_.setCalculator(calc);
    }

    //
    // IDL:ooc.com/LoadBalancing/LoadUpdater/register_member:1.0
    //
    public void register_member(String member_id, String group_id)
        throws GroupNotFound
    {
	updater_.registerMember(member_id, group_id);
    }

    //
    // IDL:ooc.com/LoadBalancing/LoadUpdater/unregister_member:1.0
    //
    public void unregister_member(String member_id, String group_id)
    {
	updater_.unregisterMember(member_id, group_id);
    }
}

class LoadUpdaterThread extends Thread
{
    private static class MemberElement
    {
	public String memberId;
	public String groupId;
	public StrategyProxy proxy;
    }

    //
    // The logger
    //
    private Logger logger_;

    //
    // The GroupFactory
    //
    private GroupFactory factory_;

    //
    // The LoadCalculator
    //
    private LoadCalculator calc_;

    //
    // The member data
    //
    private java.util.Vector members_ = new java.util.Vector();

    //
    // The update frequency
    //
    private int freq_;

    //
    // Was the update frequency changed?
    //
    private boolean freqChanged_;

    //
    // Are we finished?
    //
    private boolean done_;

    // ------------------------------------------------------------------
    // LoadUpdaterThread private member implementation
    // ------------------------------------------------------------------

    //
    // Push load to the StrategyProxy.
    //
    private void pushLoad()
    {
	if(calc_ == null)
	{
	    logger_.warning("LoadCalculator is not set");
	    return;
	}

	int load;
	try
	{
	    load = calc_.calculate_load();
	}
	catch(Exception ex)
	{
	    logger_.warning("caught " + ex.toString() +
			    " while pulling load from LoadCalculator");
	    return;
	}

	if(logger_.traceLevel() >= 2)
	{
	    String msg = "load of " + load;

	    if(logger_.traceLevel() >= 3)
	    {
		java.text.SimpleDateFormat date =
		    new java.text.SimpleDateFormat("HH:mm:ss");

		msg = msg + " at " +
		    date.format(new java.util.Date(
			System.currentTimeMillis()));
	    }

	    logger_.trace("updater", msg);
	}

	for(int i = 0 ; i < members_.size() ; i++)
	{
	    MemberElement el = (MemberElement)members_.elementAt(i);
	    try
	    {
		el.proxy.push_load(el.memberId, load);
	    }
	    catch(Exception ex)
	    {
		logger_.warning("caught " + ex.toString() +
				" while pushing load to StrategyProxy");
		return;
	    }
	}
    }

    // ------------------------------------------------------------------
    // LoadUpdaterThread constructor
    // ------------------------------------------------------------------

    public LoadUpdaterThread(
	Logger logger,
	GroupFactory factory,
	int freq)
    {
	logger_ = logger;
	factory_ = factory;
	freq_ = freq;
	freqChanged_ = false;
	done_ = false;
    }

    // ------------------------------------------------------------------
    // LoadUpdaterThread public member implementation
    // ------------------------------------------------------------------

    //
    // Get the update frequency.
    //
    public synchronized int
    getFrequency()
    {
	return freq_;
    }

    //
    // Set the update frequency.
    //
    public synchronized void
    setFrequency(int freq)
    {
	freqChanged_ = true;
	freq_ = freq;
	notify();
    }

    //
    // Set the LoadCalculator.
    //
    public synchronized void
    setCalculator(LoadCalculator calc)
    {
	calc_ = calc;
    }

    //
    // Register a member.
    //
    public synchronized void
    registerMember(String member_id, String group_id)
        throws GroupNotFound
    {
	StrategyProxy proxy = null;
	for(int i = 0 ; i < members_.size() ; i++)
	{
	    MemberElement el = (MemberElement)members_.elementAt(i);

	    if(!el.groupId.equals(group_id))
		continue; // group doesn't match

	    if(el.memberId.equals(member_id))
		return; // already registered

	    if(proxy == null)
		proxy = el.proxy;
	}

	//
	// Get the strategy proxy for the group, if it was not found above.
	//
	if(proxy == null)
	{
	    Group group = factory_.get(group_id);
	    proxy = group.get_strategy_proxy();
	}

	//
	// Add the member element.
	//
	MemberElement el = new MemberElement();
	el.proxy = proxy;
	el.memberId = member_id;
	el.groupId = group_id;
	members_.addElement(el);
    }

    //
    // Unregister a member.
    //
    public synchronized void
    unregisterMember(String member_id, String group_id)
    {
	for(int i = 0 ; i < members_.size() ; i++)
	{
	    MemberElement el = (MemberElement)members_.elementAt(i);

	    if(el.groupId.equals(group_id) &&
	       el.memberId.equals(member_id))
	    {
		members_.removeElementAt(i);
		return;
	    }
	}
    }

    //
    // Destroy the thread.
    //
    public void destroy()
    {
	done_ = true;
	notify();
    }

    //
    // Run the thread.
    //
    public synchronized void run()
    {
	while(!done_)
	{
	    try
	    {
		if(freq_ == 0)
		    wait();
		else
		    wait(freq_);
	    }
	    catch(InterruptedException ex)
	    {
	    }

	    //
	    // Are we done?
	    //
	    if(done_)
		break;

	    //
	    // Were we notified because the update frequency changed?
	    //
	    if(freqChanged_)
	    {
		freqChanged_ = false;
		continue;
	    }

	    //
	    // No updates when the frequency is 0.
	    //
	    if(freq_ == 0)
		continue;

	    //
	    // Push the load to the StrategyProxy.
	    //
	    pushLoad();
	}
    }
}
