// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OCI.UDP;

final class ServerTransport_impl extends org.omg.CORBA.LocalObject
    implements com.ooc.OCI.Transport
{
    //
    // The client ID
    //
    private int clientID_;

    //
    // The remote address
    //
    // NOTE: Must not be private, because ServerTransportInfo_impl
    // accesses this member
    //
    java.net.InetAddress remoteAddr_;

    //
    // The remote port
    //
    // NOTE: Must not be private, because ServerTransportInfo_impl
    // accesses this member
    //
    int remotePort_;

    //
    // The packet handler
    //
    // NOTE: Must not be private, because ServerTransportInfo_impl
    // accesses this member
    //
    ServerPacketHandler serverPacketHandler_;

    //
    // The trace level
    //
    private int trace_;

    //
    // The message timeout
    //
    private int messageTimeout_;

    //
    // The transport timeout
    //
    private int transportTimeout_;

    //
    // The logger
    //
    private com.ooc.OB.Logger logger_;

    //
    // The transport information object
    //
    private ServerTransportInfo_impl info_;

    //
    // Complete pending messages
    //
    private java.util.Vector messages_ = new java.util.Vector();

    //
    // The packet collector
    //
    private PacketCollector packetCollector_;

    //
    // Has this transport been shutdown?
    //
    private boolean shutdown_;

    //
    // Has this transport been closed?
    //
    private boolean closed_;

    //
    // The time we last received a packet (in seconds)
    //
    private long lastPacketTime_;

    // ------------------------------------------------------------------
    // Private and protected member implementations
    // ------------------------------------------------------------------

    private void
    fillBufferNoSync(com.ooc.OCI.Buffer buf)
    {
        //
        // Synchronization must be done by caller
        //

        while(!buf.is_full() && messages_.size() > 0)
        {
            com.ooc.OCI.Buffer message = 
                (com.ooc.OCI.Buffer )messages_.firstElement();

            com.ooc.OB.Assert._OB_assert(message.rest_length() >= 
                                     buf.rest_length());

            System.arraycopy(message.data(), message.pos(),
                             buf.data(), buf.pos(),
                             buf.rest_length());

            message.advance(buf.rest_length());
            buf.advance(buf.rest_length());

            if(message.is_full())
            {
                messages_.removeElementAt(0);
                notify();
            }
        }
    }

    private void
    receiveNoSync(com.ooc.OCI.Buffer buf, int timeout)
    {
        //
        // Synchronization must be done by caller
        //

        int delay;
        if(timeout > 0 && timeout < 5)
            delay = timeout * 1000;
        else
            delay = 5 * 1000;

        long start = System.currentTimeMillis();
        long lastSweep = start / 1000;

        long deadline = 0;
        if(timeout > 0)
            deadline = start + timeout * 1000;

        while(!buf.is_full())
        {
            if(messages_.size() == 0 && !shutdown_)
            {
                try
                {
                    //
                    // Wait until the timer expires or we are notified
                    //
                    wait(delay);
                }
                catch(InterruptedException ex)
                {
                }
            }

            fillBufferNoSync(buf);

            if(!shutdown_)
            {
                final long now = System.currentTimeMillis();

                if(now >= start + delay)
                {
                    final long nowS = now / 1000;

                    if(trace_ > 0)
                        logger_.trace("UDP", "running garbage collector");

                    if(nowS - lastPacketTime_ > transportTimeout_)
                    {
                        if(trace_ > 0)
                            logger_.trace("UDP", "reclaiming transport");

                        shutdown();
                    }
                    else if(nowS - lastSweep > messageTimeout_)
                    {
                        packetCollector_.sweep();
                        lastSweep = nowS;
                    }

                    start = now;
                }

                if(timeout > 0 && now >= deadline)
                    break;
            }

            if(shutdown_)
                break;
        }
    }

    // ------------------------------------------------------------------
    // Package member implementations
    // ------------------------------------------------------------------

    boolean
    addPacket(com.ooc.CORBA.InputStream in, int packetSize,
              com.ooc.UDPIOP.FixedPacketHeader fHeader)
    {
        com.ooc.OCI.Buffer buf = null;
        synchronized(this)
        {
            if(shutdown_)
                return false;

            lastPacketTime_ = System.currentTimeMillis() / 1000;

            //
            // Give packet to packet collector. If this packet completes
            // the message, the packet collector will return a buffer.
            //
            buf = packetCollector_.addPacket(in, packetSize, fHeader);

            //
            // If the message is complete, add it to messages_ and notify
            // the thread that may be waiting in receive_detect
            //
            if(buf != null)
            {
                messages_.addElement(buf);
                notify();
            }
        }

        return true;
    }

    // ------------------------------------------------------------------
    // Standard IDL to Java Mapping
    // ------------------------------------------------------------------

    public String
    id()
    {
        return PLUGIN_ID.value;
    }

    public int
    tag()
    {
        return com.ooc.UDPIOP.TAG_UDP_IOP.value;
    }

    public com.ooc.OCI.SendReceiveMode
    mode()
    {
        return com.ooc.OCI.SendReceiveMode.ReceiveOnly;
    }

    public int
    handle()
    {      
        throw new org.omg.CORBA.NO_IMPLEMENT();
    }

    public synchronized void
    close()
    {
        if(!closed_)
        {
            closed_ = true;

            if(!shutdown_)
                shutdown();

            //
            // Call callbacks
            //
            info_._OB_callCloseCB(info_);

            //
            // Destroy the info object
            //
            info_._OB_destroy();
        }
    }

    public synchronized void
    shutdown()
    {
        if(!shutdown_)
        {
            shutdown_ = true;

            //
            // Remove ourselves from the packet handler so that no new
            // packets are delivered to this transport
            //
            serverPacketHandler_.removeTransport(clientID_, remoteAddr_,
                                                 remotePort_);

            //
            // Notify receiver thread
            //
            notify();

            //
            // Wait for all pending messages to be received
            //
            while(messages_.size() > 0)
            {
                try
                {
                    wait();
                }
                catch(InterruptedException ex)
                {
                }
            }
        }
    }

    public void
    receive(com.ooc.OCI.Buffer buf, boolean block)
    {
        com.ooc.OB.Assert._OB_assert(false); // Should not be called by the server
    }

    public synchronized boolean
    receive_detect(com.ooc.OCI.Buffer buf, boolean block)
    {
        if(!shutdown_)
        {
            if(block)
                receiveNoSync(buf, 0);
            else
                fillBufferNoSync(buf);
        }

        return !shutdown_;
    }

    public synchronized void
    receive_timeout(com.ooc.OCI.Buffer buf, int t)
    {
        if(!shutdown_)
            receiveNoSync(buf, t);

        if(shutdown_)
            throw new org.omg.CORBA.COMM_FAILURE();
    }

    public synchronized boolean
    receive_timeout_detect(com.ooc.OCI.Buffer buf, int t)
    {
        if(!shutdown_)
            receiveNoSync(buf, t);

        return !shutdown_;
    }

    public void
    send(com.ooc.OCI.Buffer buf, boolean block)
    {
        send_detect(buf, block);
    }

    public boolean
    send_detect(com.ooc.OCI.Buffer buf, boolean block)
    {
        //
        // Dummy send
        //
        buf.advance(buf.rest_length());

        if(shutdown_)
            return false;
        else
            return true;
    }

    public void
    send_timeout(com.ooc.OCI.Buffer buf, int t)
    {
        throw new InternalError();
    }

    public boolean
    send_timeout_detect(com.ooc.OCI.Buffer buf, int t)
    {
        throw new InternalError();
    }

    public com.ooc.OCI.TransportInfo
    get_info()
    {
        return info_;
    }

    // ------------------------------------------------------------------
    // ORBacus internal functions
    // Application programs must not use these functions directly
    // ------------------------------------------------------------------

    ServerTransport_impl(com.ooc.OCI.AcceptorInfo acceptorInfo,
                         int clientID, java.net.InetAddress remoteAddr,
                         int remotePort, ServerPacketHandler s, int trace,
                         int messageTimeout, int transportTimeout,
                         com.ooc.OB.Logger logger)
    {
        clientID_ = clientID;
        remoteAddr_ = remoteAddr;
        remotePort_ = remotePort;
        serverPacketHandler_ = s;
        trace_ = trace;
        messageTimeout_ = messageTimeout;
        transportTimeout_ = transportTimeout;
        logger_ = logger;
        info_ = new ServerTransportInfo_impl(this, acceptorInfo);
        packetCollector_ = new PacketCollector(trace, messageTimeout, logger);
        shutdown_ = false;
        closed_ = false;
    }
}
